/*
 *   Copyright (C) 2009,2010 by Jonathan Naylor G4KLX
 *
 *   This program is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; either version 2 of the License, or
 *   (at your option) any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program; if not, write to the Free Software
 *   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include "AnalogueRepeaterD.h"
#include "AnalogueRepeaterConfig.h"
#include "ExternalController.h"
#include "SerialController.h"
#include "URIUSBController.h"
#include "DummyController.h"
#include "K8055Controller.h"
#include "Version.h"
#include "Logger.h"

#include <wx/cmdline.h>
#include <wx/wfstream.h>
#include <wx/fileconf.h>
#include <wx/filename.h>

const wxChar* NAME_PARAM       = wxT("Repeater Name");
const wxChar* NOLOGGING_SWITCH = wxT("nolog");

int main(int argc, char** argv)
{
	bool res = ::wxInitialize();
	if (!res) {
		::fprintf(stderr, "analoguerepeaterd: failed to initialise the wxWidgets library, exiting\n");
		return -1;
	}

	wxCmdLineParser parser(argc, argv);
	parser.AddSwitch(NOLOGGING_SWITCH, wxEmptyString, wxEmptyString, wxCMD_LINE_PARAM_OPTIONAL);
	parser.AddParam(NAME_PARAM, wxCMD_LINE_VAL_STRING, wxCMD_LINE_PARAM_OPTIONAL);

	int cmd = parser.Parse();
	if (cmd != 0) {
		::wxUninitialize();
		return 0;
	}

	bool nolog = parser.Found(NOLOGGING_SWITCH);

	wxString name;
	if (parser.GetParamCount() > 0U)
		name = parser.GetParam(0U);

	CAnalogueRepeaterD repeater(nolog, name);

	if (!repeater.init()) {
		::wxUninitialize();
		return 1;
	}

	repeater.run();

	::wxUninitialize();
	return 0;
}

CAnalogueRepeaterD::CAnalogueRepeaterD(bool nolog, const wxString& name) :
m_name(name),
m_nolog(nolog),
m_thread(NULL)
{
}

CAnalogueRepeaterD::~CAnalogueRepeaterD()
{
}

bool CAnalogueRepeaterD::init()
{
	if (!m_nolog) {
		wxString logBaseName = LOG_BASE_NAME;
		if (!m_name.IsEmpty()) {
			logBaseName.Append(wxT("_"));
			logBaseName.Append(m_name);
		}

		wxFileName loggerName(wxFileName::GetHomeDir(), logBaseName, wxT("log"));
		wxLog* log = new CLogger(loggerName.GetFullPath());
		wxLog::SetActiveTarget(log);
	} else {
		wxLog::SetActiveTarget(new wxLogStderr);
	}

	::wxLogInfo(wxT("Starting ") + APPLICATION_NAME + wxT(" - ") + VERSION);

	return createThread();
}

void CAnalogueRepeaterD::run()
{
	m_thread->Create();
	m_thread->Run();

	m_thread->Wait();
	delete m_thread;

	::wxLogInfo(APPLICATION_NAME + wxT(" is exiting"));
}

bool CAnalogueRepeaterD::createThread()
{
	wxFileName fileName(wxFileName::GetHomeDir(), CONFIG_FILE_NAME);
	wxFileInputStream stream(fileName.GetFullPath());
	if (!stream.IsOk()) {
		::wxLogError(wxT("Cannot open the config file - %s"), fileName.GetFullPath().c_str());
		return false;
	}

	CAnalogueRepeaterConfig config(new wxFileConfig(stream), m_name);

	m_thread = new CAnalogueRepeaterThread;

	wxString callsign;
	wxString beacon;
	unsigned int speed;
	unsigned int freq;
	wxFloat32 level1;
	wxFloat32 level2;
	config.getCallsign(callsign, beacon, speed, freq, level1, level2);
	m_thread->setCallsign(callsign, beacon, speed, freq, level1, level2);
	::wxLogInfo(wxT("Callsign: %s, beacon: %s, speed: %u WPM, freq: %u Hz, level1: %.3f, level2: %.3f"), callsign.c_str(), beacon.c_str(), speed, freq, level1, level2);

	wxString radioAck, extAck;
	unsigned int ack;
	unsigned int minimum;
	wxFloat32 level;
	config.getAck(radioAck, extAck, speed, freq, level, ack, minimum);
	m_thread->setAck(radioAck, extAck, speed, freq, level, ack, minimum);
	::wxLogInfo(wxT("Radio ack set to '%s', network ack set to '%s', speed: %u WPM, freq: %u Hz, level: %.3f, ack: %u secs, minimum: %u secs"), radioAck.c_str(), extAck.c_str(), speed, freq, level, ack, minimum);

	unsigned int callsignTime;
	unsigned int timeout;
	unsigned int lockoutTime;
	unsigned int hangTime;
	unsigned int latchTime;
	config.getTimes(callsignTime, timeout, lockoutTime, hangTime, latchTime);
	m_thread->setTimes(callsignTime, timeout, lockoutTime, hangTime, latchTime);
	::wxLogInfo(wxT("Times set to: callsign time: %u secs, timeout: %u secs, lockout time %u secs, hang time: %u secs, latch time: %u secs"), callsignTime, timeout, lockoutTime, hangTime, latchTime);

	bool      tbEnable;
	wxFloat32 tbThreshold;
	wxFloat32 ctcssFreq;
	bool      ctcssInternal;
	wxFloat32 ctcssThresh;
	wxFloat32 ctcssLevel;
	ANALOGUE_CTCSS_OUTPUT ctcssOutput;
	config.getTones(tbEnable, tbThreshold, ctcssFreq, ctcssInternal, ctcssThresh, ctcssLevel, ctcssOutput);
	m_thread->setTones(tbEnable, tbThreshold, ctcssFreq, ctcssInternal, ctcssThresh, ctcssLevel, ctcssOutput);
	::wxLogInfo(wxT("Tones set to: toneburst enable: %u, threshold: %.3f, CTCSS freq: %.1f Hz, internal: %u, threshold: %.3f, level: %.3f, output: %d"), tbEnable, tbThreshold, ctcssFreq, ctcssInternal, ctcssThresh, ctcssLevel, ctcssOutput);

	ANALOGUE_CALLSIGN_START callsignAtStart;
	bool callsignAtEnd;
	ANALOGUE_TIMEOUT_TYPE timeoutType;
	ANALOGUE_CALLSIGN_HOLDOFF callsignHoldoff;
	config.getFeel(callsignAtStart, callsignAtEnd, timeoutType, callsignHoldoff);
	m_thread->setFeel(callsignAtStart, callsignAtEnd, timeoutType, callsignHoldoff);
	::wxLogInfo(wxT("Feel set to: callsignAtStart: %d, callsignAtEnd: %u, timeoutType: %d, callsignHoldoff: %d"), callsignAtStart, callsignAtEnd, timeoutType, callsignHoldoff);

	wxString readDevice, writeDevice;
	unsigned int audioDelay;
	bool deEmphasis, preEmphasis, vogad;
	config.getRadio(readDevice, writeDevice, audioDelay, deEmphasis, preEmphasis, vogad);
	::wxLogInfo(wxT("Radio soundcard set to %s:%s, delay: %u ms, de-emphasis: %d, pre-emphasis %d, vogad: %d"), readDevice.c_str(), writeDevice.c_str(), audioDelay * 20U, deEmphasis, preEmphasis, vogad);

	if (!readDevice.IsEmpty() && !writeDevice.IsEmpty()) {
#if defined(__WINDOWS__)
		CSoundCardReaderWriter* soundcard = new CSoundCardReaderWriter(readDevice, writeDevice, ANALOGUE_RADIO_SAMPLE_RATE, ANALOGUE_RADIO_BLOCK_SIZE);
#else
		CSoundCardReaderWriter* soundcard = new CSoundCardReaderWriter(readDevice, writeDevice, ANALOGUE_RADIO_SAMPLE_RATE, 64U);
#endif
		soundcard->setCallback(m_thread, SOUNDCARD_RADIO);

		bool res = soundcard->open();
		if (!res) {
			::wxLogError(wxT("Cannot open the radio sound card"));
			return false;
		}

		m_thread->setRadio(soundcard, audioDelay, deEmphasis, preEmphasis, vogad);
	}

	wxString type;
	unsigned int pttDelay;
	bool pttInvert, squelchInvert;
	config.getController(type, pttDelay, pttInvert, squelchInvert);
	::wxLogInfo(wxT("Controller set to %s, ptt delay: %u ms, PTT Inversion set to %d, Squelch Inversion set to %d"), type.c_str(), pttDelay * 20U, pttInvert, squelchInvert);

	CExternalController* controller = NULL;

	wxString port;
	if (type.StartsWith(wxT("Velleman K8055 - "), &port)) {
		unsigned long num;
		port.ToULong(&num);
		controller = new CExternalController(new CK8055Controller(num), pttInvert, squelchInvert);
	} else if (type.StartsWith(wxT("URI USB - "), &port)) {
		unsigned long num;
		port.ToULong(&num);
		controller = new CExternalController(new CURIUSBController(num), pttInvert, squelchInvert);
	} else if (type.StartsWith(wxT("Serial - "), &port)) {
		controller = new CExternalController(new CSerialController(port), pttInvert, squelchInvert);
	} else {
		controller = new CExternalController(new CDummyController, pttInvert, squelchInvert);
	}

	bool res = controller->open();
	if (!res) {
		::wxLogError(wxT("Cannot open the hardware interface - %s"), type.c_str());
		return false;
	}

	m_thread->setController(controller, pttDelay);

	ANALOGUE_EXTERNAL_MODE mode;
	wxString device;
	SERIALPIN txPin, rxPin;
	config.getExternal(mode, readDevice, writeDevice, vogad, device, txPin, rxPin);
	::wxLogInfo(wxT("External mode: %d, soundcard set to %s:%s, vogad: %u, interface set to %s, tx %d, rx %d"), mode, readDevice.c_str(), writeDevice.c_str(), vogad, device.c_str(), txPin, rxPin);

	if (mode != AEM_DISABLED) {
#if defined(__WINDOWS__)
		CSoundCardReaderWriter* soundcard = new CSoundCardReaderWriter(readDevice, writeDevice, ANALOGUE_RADIO_SAMPLE_RATE, ANALOGUE_RADIO_BLOCK_SIZE);
#else
		CSoundCardReaderWriter* soundcard = new CSoundCardReaderWriter(readDevice, writeDevice, ANALOGUE_RADIO_SAMPLE_RATE, 64U);
#endif
		soundcard->setCallback(m_thread, SOUNDCARD_EXTERNAL);

		bool res = soundcard->open();
		if (!res) {
			::wxLogError(wxT("Cannot open the external sound card"));
			return false;
		}

		// This works even for an empty device name
		CNetworkController* controller = new CNetworkController(device, txPin, rxPin);

		res = controller->open();
		if (!res) {
			::wxLogError(wxT("Cannot open serial port - %s"), device.c_str());
			return false;
		}

		m_thread->setExternal(mode, soundcard, vogad, controller);
	}

	bool out1, out2, out3, out4;
	config.getOutputs(out1, out2, out3, out4);
	m_thread->setOutputs(out1, out2, out3, out4);
	::wxLogInfo(wxT("Output 1: %d, output 2: %d, output 3: %d, output 4: %d"), out1, out2, out3, out4);

	bool dtmfRadio, dtmfExternal;
	wxString dtmfShutdown, dtmfStartup, dtmfTimeout, dtmfTimeReset, dtmfOutput1, dtmfOutput2, dtmfOutput3, dtmfOutput4;
	wxString dtmfCommand1, dtmfCommand1Line, dtmfCommand2, dtmfCommand2Line;
	wxFloat32 dtmfThreshold;
	config.getDTMF(dtmfRadio, dtmfExternal, dtmfShutdown, dtmfStartup, dtmfTimeout, dtmfTimeReset, dtmfCommand1, dtmfCommand1Line, dtmfCommand2, dtmfCommand2Line, dtmfOutput1, dtmfOutput2, dtmfOutput3, dtmfOutput4, dtmfThreshold);
	m_thread->setDTMF(dtmfRadio, dtmfExternal, dtmfShutdown, dtmfStartup, dtmfTimeout, dtmfTimeReset, dtmfCommand1, dtmfCommand1Line, dtmfCommand2, dtmfCommand2Line, dtmfOutput1, dtmfOutput2, dtmfOutput3, dtmfOutput4, dtmfThreshold);
	::wxLogInfo(wxT("DTMF: Radio: %d, External: %d, Shutdown: %s, Startup: %s, Timeout: %s, Time Reset: %s, Command1: %s = %s, Command2: %s = %s, Output1: %s, Output2: %s, Output3: %s, Output4: %s, Threshold: %f"), dtmfRadio, dtmfExternal, dtmfShutdown.c_str(), dtmfStartup.c_str(), dtmfTimeout.c_str(), dtmfTimeReset.c_str(), dtmfCommand1.c_str(), dtmfCommand1Line.c_str(), dtmfCommand2.c_str(), dtmfCommand2Line.c_str(), dtmfOutput1.c_str(), dtmfOutput2.c_str(), dtmfOutput3.c_str(), dtmfOutput4.c_str(), dtmfThreshold);

	unsigned int activeHangTime;
	config.getActiveHang(activeHangTime);
	m_thread->setActiveHang(activeHangTime);
	::wxLogInfo(wxT("Active Hang: time: %u"), activeHangTime);

	m_thread->Create();
	m_thread->SetPriority(100U);
	m_thread->Run();

	return true;
}
