/*
 *   Copyright (C) 2009,2010 by Jonathan Naylor G4KLX
 *
 *   This program is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; either version 2 of the License, or
 *   (at your option) any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program; if not, write to the Free Software
 *   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include "AnalogueRepeaterFrame.h"
#include "AnalogueRepeaterApp.h"
#include "LogEvent.h"
#include "Version.h"

#include <wx/aboutdlg.h>

#include <cmath>

DEFINE_EVENT_TYPE(LOG_EVENT)

enum {
	Menu_View_Updates = 6000,

	Menu_Action_Shutdown,
	Menu_Action_Startup,
	Menu_Action_Timeout,
	Menu_Action_TimeReset,
	Menu_Action_Command1,
	Menu_Action_Command2,

	Menu_Output_1,
	Menu_Output_2,
	Menu_Output_3,
	Menu_Output_4
};

BEGIN_EVENT_TABLE(CAnalogueRepeaterFrame, wxFrame)
#if defined(__WXDEBUG__)
	EVT_MENU(wxID_OPEN, CAnalogueRepeaterFrame::onOpen)
#endif
	EVT_MENU(wxID_EXIT, CAnalogueRepeaterFrame::onQuit)
	EVT_MENU(Menu_View_Updates, CAnalogueRepeaterFrame::onUpdates)
	EVT_MENU(wxID_ABOUT, CAnalogueRepeaterFrame::onAbout)

	EVT_MENU(Menu_Action_Startup,   CAnalogueRepeaterFrame::onActions)
	EVT_MENU(Menu_Action_Shutdown,  CAnalogueRepeaterFrame::onActions)
	EVT_MENU(Menu_Action_Timeout,   CAnalogueRepeaterFrame::onActions)
	EVT_MENU(Menu_Action_TimeReset, CAnalogueRepeaterFrame::onActions)
	EVT_MENU(Menu_Action_Command1,  CAnalogueRepeaterFrame::onActions)
	EVT_MENU(Menu_Action_Command2,  CAnalogueRepeaterFrame::onActions)

	EVT_MENU(Menu_Output_1, CAnalogueRepeaterFrame::onOutputs)
	EVT_MENU(Menu_Output_2, CAnalogueRepeaterFrame::onOutputs)
	EVT_MENU(Menu_Output_3, CAnalogueRepeaterFrame::onOutputs)
	EVT_MENU(Menu_Output_4, CAnalogueRepeaterFrame::onOutputs)

	EVT_CLOSE(CAnalogueRepeaterFrame::onClose)

	EVT_TIMER(-1, CAnalogueRepeaterFrame::onTimer)

	EVT_CUSTOM(LOG_EVENT, wxID_ANY, CAnalogueRepeaterFrame::onLog)
END_EVENT_TABLE()

CAnalogueRepeaterFrame::CAnalogueRepeaterFrame(const wxString& title) :
wxFrame(NULL, -1, title),
m_timer(this),
m_panel(NULL),
m_outputMenu(NULL),
#if defined(__WXDEBUG__)
m_updates(true)
#else
m_updates(false)
#endif
{
	SetMenuBar(createMenuBar());

	wxBoxSizer* mainSizer = new wxBoxSizer(wxVERTICAL);

	wxPanel* panel = new wxPanel(this);

	wxBoxSizer* panelSizer = new wxBoxSizer(wxVERTICAL);

	m_panel = new CAnalogueRepeaterPanel(panel);
	panelSizer->Add(m_panel);

	panel->SetSizer(panelSizer);
	panelSizer->SetSizeHints(panel);

	mainSizer->Add(panel);

	SetSizer(mainSizer);
	mainSizer->SetSizeHints(this);

	m_timer.Start(1000);
}

CAnalogueRepeaterFrame::~CAnalogueRepeaterFrame()
{
}

wxMenuBar* CAnalogueRepeaterFrame::createMenuBar()
{
	wxMenu* fileMenu = new wxMenu();
#if defined(__WXDEBUG__)
	fileMenu->Append(wxID_OPEN, _("Open...\tCtrl+N"));
	fileMenu->AppendSeparator();
#endif
	fileMenu->Append(wxID_EXIT, _("Exit"));

	wxMenu* viewMenu = new wxMenu();
	viewMenu->AppendCheckItem(Menu_View_Updates, _("GUI Updates"));
	viewMenu->Check(Menu_View_Updates, m_updates);

	wxMenu* actionMenu = new wxMenu();
	actionMenu->Append(Menu_Action_Shutdown,  _("Shutdown"));
	actionMenu->Append(Menu_Action_Startup,   _("Startup"));
	actionMenu->Append(Menu_Action_Timeout,   _("Timeout"));
	actionMenu->Append(Menu_Action_TimeReset, _("Time Reset"));

	m_outputMenu = new wxMenu();
	m_outputMenu->AppendCheckItem(Menu_Output_1, _("Output 1"));
	m_outputMenu->AppendCheckItem(Menu_Output_2, _("Output 2"));
	m_outputMenu->AppendCheckItem(Menu_Output_3, _("Output 3"));
	m_outputMenu->AppendCheckItem(Menu_Output_4, _("Output 4"));

	wxMenu* helpMenu = new wxMenu();
	helpMenu->Append(wxID_ABOUT, _("About Analogue Repeater"));

	wxMenuBar* menuBar = new wxMenuBar();
	menuBar->Append(fileMenu,     _("File"));
	menuBar->Append(viewMenu,     _("View"));
	menuBar->Append(actionMenu,   _("Action"));
	menuBar->Append(m_outputMenu, _("Output"));
	menuBar->Append(helpMenu,     _("Help"));

	return menuBar;
}

#if defined(__WXDEBUG__)
void CAnalogueRepeaterFrame::onOpen(wxCommandEvent& event)
{
	wxFileDialog dialog(this, _("Select a WAV File"), wxEmptyString, wxEmptyString, _("WAV file (*.wav)|*.wav|All files (*.*)|*.*"));
	if (dialog.ShowModal() != wxID_OK)
		return;

	wxString fileName = dialog.GetPath();

	CWAVFileReader* reader = new CWAVFileReader(fileName, ANALOGUE_RADIO_BLOCK_SIZE);
	bool res = reader->open();
	if (!res) {
		wxMessageDialog dialog(this, _("Cannot find the file, or the file is corrupt."), _("Analogue Repeater Error"), wxICON_ERROR);
		dialog.ShowModal();
		delete reader;
		return;
	}

	if (reader->getSampleRate() != ANALOGUE_RADIO_SAMPLE_RATE) {
		::wxLogError(wxT("File %s has an invalid sample rate of %u"), fileName.c_str(), reader->getSampleRate());
		wxMessageDialog dialog(this, _("The sound file has an invalid sample rate."), _("Analogue Repeater Error"), wxICON_ERROR);
		dialog.ShowModal();
		reader->close();
		delete reader;
		return;
	}

	if (reader->getChannels() != 1U) {
		::wxLogError(wxT("File %s has too many channels - %u"), fileName.c_str(), reader->getChannels());
		wxMessageDialog dialog(this, _("The sound file has too many channels."), _("Analogue Repeater Error"), wxICON_ERROR);
		dialog.ShowModal();
		reader->close();
		delete reader;
		return;
	}

	::wxGetApp().setReader(reader);
}
#endif

void CAnalogueRepeaterFrame::onQuit(wxCommandEvent& event)
{
	Close(false);
}

void CAnalogueRepeaterFrame::onClose(wxCloseEvent& event)
{
	Destroy();
}

void CAnalogueRepeaterFrame::onUpdates(wxCommandEvent& event)
{
	m_updates = event.IsChecked();
}

void CAnalogueRepeaterFrame::onAbout(wxCommandEvent& event)
{
	wxAboutDialogInfo info;
	info.AddDeveloper(wxT("Jonathan Naylor, G4KLX"));
	info.SetCopyright(wxT("(C) 2009,2010 using GPL v2 or later"));
	info.SetName(APPLICATION_NAME);
	info.SetVersion(VERSION);
	info.SetDescription(_("This program allows a computer with a soundcard to act as a repeater."));

	::wxAboutBox(info);
}

void CAnalogueRepeaterFrame::onActions(wxCommandEvent& event)
{
	int id = event.GetId();

	switch (id) {
		case Menu_Action_Shutdown:
			::wxGetApp().shutdown();
			break;
		case Menu_Action_Startup:
			::wxGetApp().startup();
			break;
		case Menu_Action_Timeout:
			::wxGetApp().timeout();
			break;
		case Menu_Action_TimeReset:
			::wxGetApp().timeReset();
			break;
		case Menu_Action_Command1:
			::wxGetApp().command1();
			break;
		case Menu_Action_Command2:
			::wxGetApp().command2();
			break;
	}
}

void CAnalogueRepeaterFrame::onOutputs(wxCommandEvent& event)
{
	bool out1 = m_outputMenu->IsChecked(Menu_Output_1);
	bool out2 = m_outputMenu->IsChecked(Menu_Output_2);
	bool out3 = m_outputMenu->IsChecked(Menu_Output_3);
	bool out4 = m_outputMenu->IsChecked(Menu_Output_4);

	::wxGetApp().setOutputs(out1, out2, out3, out4);
}

void CAnalogueRepeaterFrame::onTimer(wxTimerEvent& event)
{
	if (!m_updates)
		return;

	CAnalogueRepeaterStatusData* status = ::wxGetApp().getStatus();
	if (status == NULL)
		return;

	m_panel->update(*status);

	delete status;
}

void CAnalogueRepeaterFrame::onLog(wxEvent& event)
{
	CLogEvent& logEvent = dynamic_cast<CLogEvent&>(event);

	wxString text = logEvent.getText();

	m_panel->showLog(text);
}

void CAnalogueRepeaterFrame::showLog(const wxString& text)
{
	if (!m_updates)
		return;

	CLogEvent event(text, LOG_EVENT);

	AddPendingEvent(event);
}

void CAnalogueRepeaterFrame::setOutputs(bool out1, bool out2, bool out3, bool out4)
{
	m_outputMenu->Check(Menu_Output_1, out1);
	m_outputMenu->Check(Menu_Output_2, out2);
	m_outputMenu->Check(Menu_Output_3, out3);
	m_outputMenu->Check(Menu_Output_4, out4);
}
