/*
 *   Copyright (C) 2009,2010 by Jonathan Naylor G4KLX
 *
 *   This program is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; either version 2 of the License, or
 *   (at your option) any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program; if not, write to the Free Software
 *   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include "DStarRepeaterStatusData.h"
#include "DStarRepeaterThread.h"
#include "RadioHeaderDecoder.h"
#include "RadioHeaderEncoder.h"
#include "DStarRepeaterApp.h"
#include "DStarDefines.h"
#include "HeaderData.h"
#include "Utils.h"

// #define	TX_TO_WAV_FILE

const unsigned int MAX_LOST_SYNCS = 3U;

CDStarRepeaterThread::CDStarRepeaterThread() :
wxThread(wxTHREAD_JOINABLE),
m_soundcard(NULL),
m_protocolHandler(NULL),
m_networkDelay(NULL),
m_controller(NULL),
m_pttDelay(NULL),
m_audioDelay(NULL),
m_stopped(true),
m_inBuffer(DSTAR_RADIO_BLOCK_SIZE * 5U),
m_outBuffer(DSTAR_RADIO_BLOCK_SIZE * 10U),
m_dataBuffer(DSTAR_GMSK_SYMBOL_RATE),		// One second of data
m_rptCallsign(),
m_gwyCallsign(),
m_reader(NULL),
m_demodulator(),
m_modulator(),
m_bitBuffer(DV_FRAME_LENGTH_BITS),
m_rxState(DSRXS_LISTENING),
m_frameMatcher(FRAME_SYNC_LENGTH_BITS, FRAME_SYNC_BITS),
m_dataMatcher(DATA_FRAME_LENGTH_BITS, DATA_SYNC_BITS),
m_endMatcher(END_PATTERN_LENGTH_BITS, END_PATTERN_BITS),
m_writer(NULL),
m_userHeader(NULL),
m_statusHeader(NULL),
m_radioSyncCounter(0U),
m_radioSyncsLost(0U),
m_networkSyncCounter(0U),
m_timeoutTimer(DSTAR_TICKS_PER_SEC, 180U),		// 180s
m_watchdogTimer(DSTAR_TICKS_PER_SEC, 0U, 500U),	// 0.5s
m_pollTimer(DSTAR_TICKS_PER_SEC, 60U),			// 60s
m_ackTimer(DSTAR_TICKS_PER_SEC, 0U, 500U),		// 0.5s
m_rptState(DSRS_LISTENING),
m_slowData(),
m_tx(false),
m_killed(false),
m_mode(MODE_DUPLEX),
m_ack(true),
m_level(1.0F),
m_controlEnabled(false),
m_controlRPT1(),
m_controlRPT2(),
m_controlShutdown(),
m_controlStartup(),
m_controlCommand1(),
m_controlCommand1Line(),
m_controlCommand2(),
m_controlCommand2Line(),
m_controlOutput1(),
m_controlOutput2(),
m_controlOutput3(),
m_controlOutput4(),
m_output1(false),
m_output2(false),
m_output3(false),
m_output4(false),
m_activeHangTimer(DSTAR_TICKS_PER_SEC),
m_shutdown(false),
m_disable(false),
m_ambe(),
m_ambeBits(0U),
m_ambeErrors(0U)
{
	setRadioState(DSRXS_LISTENING);
	setRepeaterState(DSRS_LISTENING);
}

CDStarRepeaterThread::~CDStarRepeaterThread()
{
	delete m_userHeader;
	delete m_statusHeader;
}

bool CDStarRepeaterThread::init()
{
	m_controller->setActive(false);
	m_controller->setRadioTransmit(false);

	// Send the first network poll and start the timer for the next
	if (m_protocolHandler != NULL) {
		m_protocolHandler->writePoll();
		m_pollTimer.start();
	}

	m_inBuffer.clear();

	return true;
}

void* CDStarRepeaterThread::Entry()
{
	// Wait here until we have the essentials to run
	while (!m_killed && (m_soundcard == NULL  || m_rptCallsign.IsEmpty() || m_rptCallsign.IsSameAs(wxT("        ")) || m_controller == NULL))
		Sleep(500UL);		// 1/2 sec

	if (m_killed)
		return NULL;

	m_stopped = false;

	init();

	::wxLogMessage(wxT("Starting the D-Star repeater thread"));

	unsigned int count = 0U;

	while (!m_killed) {
		// Process the incoming D-Star transmission
		receive();

		repeaterStateMachine();

		// Send the network poll if needed and restart the timer
		if (m_pollTimer.hasExpired()) {
			m_protocolHandler->writePoll();
			m_pollTimer.reset();
		}

		// Clock the heartbeat output every one second
		count++;
		if (count == 50U) {
			m_controller->setHeartbeat();
			count = 0U;
		}

		// Set the output state
		if (m_rptState == DSRS_VALID      || m_rptState == DSRS_INVALID      || m_rptState == DSRS_TIMEOUT      ||
			m_rptState == DSRS_VALID_WAIT || m_rptState == DSRS_INVALID_WAIT || m_rptState == DSRS_TIMEOUT_WAIT ||
			m_rptState == DSRS_NETWORK    || (m_activeHangTimer.isRunning() && !m_activeHangTimer.hasExpired())) {
			m_controller->setActive(true);
		} else {
			m_controller->setActive(false);
			m_activeHangTimer.stop();
		}

		// Check the shutdown state, state changes are done here to bypass the state machine which is
		// frozen when m_disable or m_shutdown are asserted
		m_disable = m_controller->getDisable();
		if (m_disable || m_shutdown) {
			if (m_rptState != DSRS_SHUTDOWN) {
				m_timeoutTimer.stop();
				m_watchdogTimer.stop();
				m_activeHangTimer.stop();
				m_ackTimer.stop();
				m_dataBuffer.clear();
				m_controller->setActive(false);
				m_controller->setRadioTransmit(false);
				m_rptState = DSRS_SHUTDOWN;
			}
		} else {
			if (m_rptState == DSRS_SHUTDOWN) {
				m_timeoutTimer.stop();
				m_watchdogTimer.stop();
				m_ackTimer.stop();
				m_rptState = DSRS_LISTENING;
				if (m_protocolHandler != NULL)	// Tell the protocol handler
					m_protocolHandler->reset();
			}
		}

		// Send the output data
		transmit();

		clock();
	}

	::wxLogMessage(wxT("Stopping the D-Star repeater thread"));

	finish();

	return NULL;
}

void CDStarRepeaterThread::finish()
{
	m_controller->setActive(false);
	m_controller->setRadioTransmit(false);
	m_controller->close();
	delete m_controller;

	m_soundcard->close();
	delete m_soundcard;

	delete m_networkDelay;
	delete m_audioDelay;
	delete m_pttDelay;

	if (m_reader != NULL) {
		m_reader->close();
		delete m_reader;
	}

	if (m_protocolHandler != NULL) {
		m_protocolHandler->close();
		delete m_protocolHandler;
	}

#if defined(TX_TO_WAV_FILE)
	if (m_writer != NULL) {
		m_writer->close();
		delete m_writer;
	}
#endif
}

void CDStarRepeaterThread::kill()
{
	m_killed = true;
}

#if defined(__WXDEBUG__)
void CDStarRepeaterThread::setReader(CWAVFileReader* reader)
{
	wxASSERT(reader != NULL);

	// Already busy?
	if (m_reader != NULL) {
		reader->close();
		delete reader;
		return;
	}

	::wxLogInfo(wxT("Reading from WAV file - %s"), reader->getFilename().c_str());

	m_reader = reader;
}
#endif

void CDStarRepeaterThread::setCallsign(const wxString& callsign, const wxString& gateway, DSTAR_MODE mode, bool ack)
{
	// Pad the callsign up to eight characters
	m_rptCallsign = callsign;
	m_rptCallsign.Append(wxT("        "));
	m_rptCallsign.Truncate(LONG_CALLSIGN_LENGTH);

	if (gateway.IsEmpty()) {
		m_gwyCallsign = callsign;
		m_gwyCallsign.Append(wxT("        "));
		m_gwyCallsign.Truncate(LONG_CALLSIGN_LENGTH - 1U);
		m_gwyCallsign.Append(wxT("G"));
	} else {
		m_gwyCallsign = gateway;
		m_gwyCallsign.Append(wxT("        "));
		m_gwyCallsign.Truncate(LONG_CALLSIGN_LENGTH);
	}

	m_mode = mode;
	m_ack  = ack;

	if (m_protocolHandler != NULL)
		m_protocolHandler->setRepeaterCallsign(m_rptCallsign);
}

void CDStarRepeaterThread::setProtocolHandler(IProtocolHandler* handler, unsigned int networkDelay)
{
	m_protocolHandler = handler;

	if (m_protocolHandler == NULL)
		return;

	bool res = m_protocolHandler->open();
	if (!res) {
		delete m_protocolHandler;
		m_protocolHandler = NULL;
	}

	m_protocolHandler->setRepeaterCallsign(m_rptCallsign);

	if (networkDelay > 0U)
		m_networkDelay = new CNetworkDelay(networkDelay);
}

void CDStarRepeaterThread::setSoundCard(CSoundCardReaderWriter* soundcard, wxFloat32 level, bool rxInvert, bool txInvert)
{
	wxASSERT(soundcard != NULL);

	m_soundcard = soundcard;
	m_level     = level;

	m_demodulator.setInvert(rxInvert);
	m_modulator.setInvert(txInvert);
}

void CDStarRepeaterThread::setTimes(unsigned int timeout, unsigned int ackTime)
{
	m_timeoutTimer.setTimeout(timeout);

	if (ackTime < 100U)		// Handle old values
		ackTime = 500U;

	m_ackTimer.setTimeout(0U, ackTime);
}

void CDStarRepeaterThread::setController(CExternalController* controller, int pttDelay)
{
	wxASSERT(controller != NULL);

	m_controller = controller;

	if (pttDelay != 0) {
		if (pttDelay > 0)
			m_pttDelay = new CPTTDelay(pttDelay);
		else
			m_audioDelay = new CAudioDelay(-pttDelay * DSTAR_FRAME_TIME_MS * DSTAR_RADIO_SAMPLE_RATE / 1000U);
	}
}

void CDStarRepeaterThread::setControl(bool enabled, const wxString& rpt1Callsign, const wxString& rpt2Callsign, const wxString& shutdown, const wxString& startup, const wxString& command1, const wxString& command1Line, const wxString& command2, const wxString& command2Line, const wxString& output1, const wxString& output2, const wxString& output3, const wxString& output4)
{
	m_controlEnabled      = enabled;
	m_controlRPT1         = rpt1Callsign;
	m_controlRPT2         = rpt2Callsign;
	m_controlShutdown     = shutdown;
	m_controlStartup      = startup;
	m_controlCommand1     = command1;
	m_controlCommand1Line = command1Line;
	m_controlCommand2     = command2;
	m_controlCommand2Line = command2Line;
	m_controlOutput1      = output1;
	m_controlOutput2      = output2;
	m_controlOutput3      = output3;
	m_controlOutput4      = output4;

	m_controlRPT1.Append(wxT(' '), LONG_CALLSIGN_LENGTH);
	m_controlRPT2.Append(wxT(' '), LONG_CALLSIGN_LENGTH);
	m_controlShutdown.Append(wxT(' '), LONG_CALLSIGN_LENGTH);
	m_controlStartup.Append(wxT(' '), LONG_CALLSIGN_LENGTH);
	m_controlCommand1.Append(wxT(' '), LONG_CALLSIGN_LENGTH);
	m_controlCommand2.Append(wxT(' '), LONG_CALLSIGN_LENGTH);
	m_controlOutput1.Append(wxT(' '), LONG_CALLSIGN_LENGTH);
	m_controlOutput2.Append(wxT(' '), LONG_CALLSIGN_LENGTH);
	m_controlOutput3.Append(wxT(' '), LONG_CALLSIGN_LENGTH);
	m_controlOutput4.Append(wxT(' '), LONG_CALLSIGN_LENGTH);

	m_controlRPT1.Truncate(LONG_CALLSIGN_LENGTH);
	m_controlRPT2.Truncate(LONG_CALLSIGN_LENGTH);
	m_controlShutdown.Truncate(LONG_CALLSIGN_LENGTH);
	m_controlStartup.Truncate(LONG_CALLSIGN_LENGTH);
	m_controlCommand1.Truncate(LONG_CALLSIGN_LENGTH);
	m_controlCommand2.Truncate(LONG_CALLSIGN_LENGTH);
	m_controlOutput1.Truncate(LONG_CALLSIGN_LENGTH);
	m_controlOutput2.Truncate(LONG_CALLSIGN_LENGTH);
	m_controlOutput3.Truncate(LONG_CALLSIGN_LENGTH);
	m_controlOutput4.Truncate(LONG_CALLSIGN_LENGTH);
}

void CDStarRepeaterThread::setActiveHang(unsigned int time)
{
	m_activeHangTimer.setTimeout(time);
}

void CDStarRepeaterThread::setOutputs(bool out1, bool out2, bool out3, bool out4)
{
	if (m_controller == NULL)
		return;

	m_output1 = out1;
	m_output2 = out2;
	m_output3 = out3;
	m_output4 = out4;

	m_controller->setOutput1(m_output1);
	m_controller->setOutput2(m_output2);
	m_controller->setOutput3(m_output3);
	m_controller->setOutput4(m_output4);
}

void CDStarRepeaterThread::receive()
{
	// Get an audio block from the RX
	wxFloat32 audio[DSTAR_RADIO_BLOCK_SIZE];

	unsigned int length = 0U;
	for (unsigned int n = 0U; length < DSTAR_RADIO_BLOCK_SIZE && n < 20U; n++) {
		length += m_inBuffer.getData(audio + length, DSTAR_RADIO_BLOCK_SIZE - length);

		if (length < DSTAR_RADIO_BLOCK_SIZE)
			Sleep(DSTAR_FRAME_TIME_MS / 4UL);
	}

	if (length < DSTAR_RADIO_BLOCK_SIZE)
		::wxLogError(wxT("No audio is being received"));

	for (unsigned int i = 0U; i < length; i++) {
		TRISTATE state = m_demodulator.decode(audio[i]);

		switch (state) {
			case STATE_TRUE:
				radioStateMachine(true);
				break;
			case STATE_FALSE:
				radioStateMachine(false);
				break;
			case STATE_UNKNOWN:
				break;
		}
	}

	if (m_protocolHandler != NULL) {
		NETWORK_TYPE type = m_protocolHandler->read();
		CHeaderData* header = NULL;
		unsigned char data[DV_FRAME_MAX_LENGTH_BYTES];
		unsigned int length = 0U;
		bool end = false;

		// Get the data from the network
		switch (type) {
			// Nothing received
			case NETWORK_NONE:
				break;

			// Either a user or a status header
			case NETWORK_USER_HEADER:
			case NETWORK_STATUS_HEADER:
				header = m_protocolHandler->readHeader();
				if (header == NULL)
					type = NETWORK_NONE;
				break;

			// Either user or status AMBE data and slow data
			case NETWORK_USER_DATA:
			case NETWORK_STATUS_DATA:
				length = m_protocolHandler->readData(data, DV_FRAME_MAX_LENGTH_BYTES, end);
				if (length == 0U)
					type = NETWORK_NONE;
				else
					m_watchdogTimer.reset();
				break;
		}

		// Have we missed a data frame?
		if (type == NETWORK_NONE && m_rptState == DSRS_NETWORK) {
			// Create a silence frame
			::memcpy(data, NULL_AMBE_DATA_BYTES, VOICE_FRAME_LENGTH_BYTES);
			::memcpy(data + VOICE_FRAME_LENGTH_BYTES, NULL_SLOW_DATA_BYTES, DATA_FRAME_LENGTH_BYTES);
			length = DV_FRAME_LENGTH_BYTES;
			end    = false;
			type   = NETWORK_USER_DATA;
		}

		// Delay the network data, if enabled
		if (m_networkDelay != NULL) {
			// Feed the data into the delay, and get the type to be read back
			switch (type) {
				case NETWORK_NONE:
					type = m_networkDelay->delay();
					break;
				case NETWORK_USER_HEADER:
				case NETWORK_STATUS_HEADER:
					type = m_networkDelay->delay(type, header);
					break;
				case NETWORK_USER_DATA:
				case NETWORK_STATUS_DATA:
					type = m_networkDelay->delay(type, data, length, end);
					break;
			}

			// Get the delayed data out
			switch (type) {
				case NETWORK_NONE:
					break;

				case NETWORK_USER_HEADER:
				case NETWORK_STATUS_HEADER:
					header = m_networkDelay->readHeader();
					if (header != NULL)
						processNetworkHeader(type, header);
					break;

				case NETWORK_USER_DATA:
				case NETWORK_STATUS_DATA:
					length = m_networkDelay->readData(data, DV_FRAME_MAX_LENGTH_BYTES, end);
					if (length > 0U)
						processNetworkFrame(type, data, length, end);
					break;
			}
		} else {
			// Dispatch the non delayed data to the repeater
			switch (type) {
				case NETWORK_NONE:
					break;

				case NETWORK_USER_HEADER:
				case NETWORK_STATUS_HEADER:
					processNetworkHeader(type, header);
					break;

				case NETWORK_USER_DATA:
				case NETWORK_STATUS_DATA:
					processNetworkFrame(type, data, length, end);
					break;
			}
		}
	}
}

void CDStarRepeaterThread::transmitHeader(CHeaderData& header)
{
	::wxLogMessage(wxT("Transmitting to - My: %s/%s  Your: %s  Rpt1: %s  Rpt2: %s  Flags: %02X %02X %02X"), header.getMyCall1().c_str(), header.getMyCall2().c_str(), header.getYourCall().c_str(), header.getRptCall1().c_str(), header.getRptCall2().c_str(), header.getFlag1(), header.getFlag2(), header.getFlag3());

	// Send the bit sync, 700 bits in 4 bit chunks
	for (unsigned int i = 0U; i < 175U; i++)
		m_dataBuffer.addData(BIT_SYNC_BITS, BIT_SYNC_LENGTH_BITS);

	// Send the frame sync
	m_dataBuffer.addData(FRAME_SYNC_BITS, FRAME_SYNC_LENGTH_BITS);

	// Send the header
	bool headerBits[FEC_SECTION_LENGTH_BITS];
	CRadioHeaderEncoder encoder(header);
	encoder.getRadioData(headerBits, FEC_SECTION_LENGTH_BITS);
	m_dataBuffer.addData(headerBits, FEC_SECTION_LENGTH_BITS);
}

void CDStarRepeaterThread::transmitStatus()
{
	if (m_statusHeader == NULL) {
		// No status received from the gateway, create our own
		CHeaderData header(m_gwyCallsign, wxT("    "), m_userHeader->getMyCall1(), m_gwyCallsign, m_rptCallsign);
		header.setRepeaterFlags(RELAY_UNAVAILABLE);
		transmitHeader(header);
	} else {
		// Status received from the gateway, remove the repeater flag from the status
		unsigned char flag = m_statusHeader->getFlag1();
		flag &= ~REPEATER_MASK;
		m_statusHeader->setFlag1(flag);

		transmitHeader(*m_statusHeader);

		// Make sure we don't transmit it again
		delete m_statusHeader;
		m_statusHeader = NULL;
	}

	// Filler data
	m_dataBuffer.addData(NULL_AMBE_DATA_BITS, VOICE_FRAME_LENGTH_BITS);
	m_dataBuffer.addData(DATA_SYNC_BITS,      DATA_FRAME_LENGTH_BITS);

	m_dataBuffer.addData(NULL_AMBE_DATA_BITS, VOICE_FRAME_LENGTH_BITS);
	m_dataBuffer.addData(NULL_SLOW_DATA_BITS, DATA_FRAME_LENGTH_BITS);

	m_dataBuffer.addData(NULL_AMBE_DATA_BITS, VOICE_FRAME_LENGTH_BITS);
	m_dataBuffer.addData(NULL_SLOW_DATA_BITS, DATA_FRAME_LENGTH_BITS);

	m_dataBuffer.addData(NULL_AMBE_DATA_BITS, VOICE_FRAME_LENGTH_BITS);
	m_dataBuffer.addData(NULL_SLOW_DATA_BITS, DATA_FRAME_LENGTH_BITS);

	m_dataBuffer.addData(NULL_AMBE_DATA_BITS, VOICE_FRAME_LENGTH_BITS);
	m_dataBuffer.addData(NULL_SLOW_DATA_BITS, DATA_FRAME_LENGTH_BITS);

	m_dataBuffer.addData(NULL_AMBE_DATA_BITS, VOICE_FRAME_LENGTH_BITS);
	m_dataBuffer.addData(NULL_SLOW_DATA_BITS, DATA_FRAME_LENGTH_BITS);

	m_dataBuffer.addData(NULL_AMBE_DATA_BITS, VOICE_FRAME_LENGTH_BITS);
	m_dataBuffer.addData(END_PATTERN_BITS,    END_PATTERN_LENGTH_BITS);

	m_dataBuffer.addData(END_PATTERN_BITS,    END_PATTERN_LENGTH_BITS);
	m_dataBuffer.addData(END_PATTERN_BITS,    END_PATTERN_LENGTH_BITS);
}

void CDStarRepeaterThread::transmitErrorStatus()
{
	CHeaderData header(m_rptCallsign, wxT("CALL"), m_userHeader->getMyCall1(), m_gwyCallsign, m_rptCallsign);
	header.setRepeaterFlags(RELAY_UNAVAILABLE);

	transmitHeader(header);

	// Filler data
	m_dataBuffer.addData(NULL_AMBE_DATA_BITS, VOICE_FRAME_LENGTH_BITS);
	m_dataBuffer.addData(DATA_SYNC_BITS,      DATA_FRAME_LENGTH_BITS);

	m_dataBuffer.addData(NULL_AMBE_DATA_BITS, VOICE_FRAME_LENGTH_BITS);
	m_dataBuffer.addData(NULL_SLOW_DATA_BITS, DATA_FRAME_LENGTH_BITS);

	m_dataBuffer.addData(NULL_AMBE_DATA_BITS, VOICE_FRAME_LENGTH_BITS);
	m_dataBuffer.addData(NULL_SLOW_DATA_BITS, DATA_FRAME_LENGTH_BITS);

	m_dataBuffer.addData(NULL_AMBE_DATA_BITS, VOICE_FRAME_LENGTH_BITS);
	m_dataBuffer.addData(NULL_SLOW_DATA_BITS, DATA_FRAME_LENGTH_BITS);

	m_dataBuffer.addData(NULL_AMBE_DATA_BITS, VOICE_FRAME_LENGTH_BITS);
	m_dataBuffer.addData(NULL_SLOW_DATA_BITS, DATA_FRAME_LENGTH_BITS);

	m_dataBuffer.addData(NULL_AMBE_DATA_BITS, VOICE_FRAME_LENGTH_BITS);
	m_dataBuffer.addData(NULL_SLOW_DATA_BITS, DATA_FRAME_LENGTH_BITS);

	m_dataBuffer.addData(NULL_AMBE_DATA_BITS, VOICE_FRAME_LENGTH_BITS);
	m_dataBuffer.addData(END_PATTERN_BITS,    END_PATTERN_LENGTH_BITS);

	m_dataBuffer.addData(END_PATTERN_BITS,    END_PATTERN_LENGTH_BITS);
	m_dataBuffer.addData(END_PATTERN_BITS,    END_PATTERN_LENGTH_BITS);
}

void CDStarRepeaterThread::transmit()
{
	if (m_dataBuffer.isEmpty() && m_outBuffer.isEmpty()) {
		// Switch off the transmitter?
		m_tx = false;

#if defined(TX_TO_WAV_FILE)
		if (m_writer != NULL) {
			::wxLogMessage(wxT("Closing the output WAV file"));
			m_writer->close();
			delete m_writer;
			m_writer = NULL;
		}
#endif
	} else if (!m_dataBuffer.isEmpty() && m_outBuffer.isEmpty()) {
		// Switch on the transmitter
		m_tx = true;

#if defined(TX_TO_WAV_FILE)
		wxDateTime time;
		time.SetToCurrent();

		// This is for logging the output audio to a WAV file
		wxString fileName = time.Format(wxT("%Y%m%d-%H%M%S.wav"));
		CWAVFileWriter* writer = new CWAVFileWriter(fileName, DSTAR_RADIO_SAMPLE_RATE, 1U, 16U, DSTAR_RADIO_BLOCK_SAMPLES);

		bool res = writer->open();
		if (!res) {
			::wxLogError(wxT("Cannot open the output WAV file - %s"), fileName.c_str());
			delete writer;
		} else {
			::wxLogMessage(wxT("Opened the output WAV file - %s"), fileName.c_str());
			m_writer = writer;
		}
#endif
		// Put quite a lot of data into the buffer to allow for slow downs
		bool bits[DV_FRAME_LENGTH_BITS * 3U];
		unsigned int length = m_dataBuffer.getData(bits, DV_FRAME_LENGTH_BITS * 3U);

		transmitBits(bits, length);
	} else if (!m_dataBuffer.isEmpty() && !m_outBuffer.isEmpty()) {
		m_tx = true;

		bool bits[DV_FRAME_LENGTH_BITS];
		unsigned int length = m_dataBuffer.getData(bits, DV_FRAME_LENGTH_BITS);

		transmitBits(bits, length);
	} else {	// m_dataBuffer.isEmpty() && !m_outBuffer.isEmpty()
		m_tx = true;
	}

	if (m_pttDelay != NULL)
		m_controller->setRadioTransmit(m_pttDelay->delay(m_tx));
	else
		m_controller->setRadioTransmit(m_tx);
}

void CDStarRepeaterThread::transmitBits(const bool* bits, unsigned int length)
{
	wxASSERT(bits != NULL);

	for (unsigned int i = 0U; i < length; i++) {
		wxFloat32 buffer[DSTAR_RADIO_BIT_LENGTH];
		unsigned int len = m_modulator.code(bits[i], buffer, DSTAR_RADIO_BIT_LENGTH);

#if defined(TX_TO_WAV_FILE)
		// If writing the transmit audio to a WAV file
		if (m_writer != NULL)
			m_writer->write(buffer, len);
#endif
		m_outBuffer.addData(buffer, len);
	}
}

void CDStarRepeaterThread::radioStateMachine(bool bit)
{
	switch (m_rxState) {
		case DSRXS_LISTENING: {
				bool res1 = m_frameMatcher.add(bit);
				bool res2 = m_dataMatcher.add(bit);

				// The frame sync has been seen
				if (res1) {
					// ::wxLogMessage(wxT("Found frame sync"));
					setRadioState(DSRXS_PROCESS_HEADER);
				}

				// The data sync has been seen
				if (res2) {
					// ::wxLogMessage(wxT("Found data sync"));
					setRadioState(DSRXS_PROCESS_SLOW_DATA);
				}
			}
			break;

		case DSRXS_PROCESS_HEADER:
			m_bitBuffer.add(bit);
			if (m_bitBuffer.getLength() == FEC_SECTION_LENGTH_BITS) {
				CHeaderData* header = processFECHeader();

				if (header != NULL) {
					bool res = processRadioHeader(header);
					if (res) {
						// A valid header and is a DV packet
						m_radioSyncCounter = 20U;
						m_radioSyncsLost   = 0U;
						setRadioState(DSRXS_PROCESS_DATA);
					} else {
						// This is a DD packet or some other problem
						// ::wxLogMessage(wxT("Invalid header"));
						setRadioState(DSRXS_LISTENING);
					}
				} else {
					// The checksum failed
					setRadioState(DSRXS_LISTENING);
				}
			}
			break;

		case DSRXS_PROCESS_DATA: {
				m_bitBuffer.add(bit);
				bool res1 = m_endMatcher.add(bit);
				bool res2 = m_dataMatcher.add(bit);

				if (res1) {
					// ::wxLogMessage(wxT("Found end sync"));
					processRadioFrame(FRAME_END);
					setRadioState(DSRXS_LISTENING);
					endOfRadioData();
					break;
				}

				if (m_bitBuffer.getLength() == DV_FRAME_LENGTH_BITS) {
					if (res2) {
						// ::wxLogMessage(wxT("Found data sync at frame %u"), m_syncCounter);
						m_radioSyncCounter = 0U;
						m_radioSyncsLost   = 0U;
						processRadioFrame(FRAME_SYNC);
					} else if (m_radioSyncCounter == 20U) {
						// ::wxLogMessage(wxT("Regenerating data sync"));
						m_radioSyncCounter = 0U;
						m_radioSyncsLost++;
						if (m_radioSyncsLost == MAX_LOST_SYNCS) {
							// ::wxLogMessage(wxT("Lost sync"));
							processRadioFrame(FRAME_END);
							setRadioState(DSRXS_LISTENING);
							endOfRadioData();
						} else {
							processRadioFrame(FRAME_SYNC);
						}
					} else {
						m_radioSyncCounter++;
						processRadioFrame(FRAME_NORMAL);
					}

					m_bitBuffer.clear();
				}
			}
			break;

		case DSRXS_PROCESS_SLOW_DATA: {
				m_bitBuffer.add(bit);
				bool res1 = m_endMatcher.add(bit);
				bool res2 = m_dataMatcher.add(bit);

   				if (res1) {
					// ::wxLogMessage(wxT("Found end sync"));
					setRadioState(DSRXS_LISTENING);
					break;
				}

				if (m_bitBuffer.getLength() == DV_FRAME_LENGTH_BITS) {
					if (res2) {
						// ::wxLogMessage(wxT("Found data sync at frame %u"), m_syncCounter);
						m_radioSyncCounter = 0U;
						m_radioSyncsLost   = 0U;
						processSlowData(true);
					} else if (m_radioSyncCounter == 20U) {
						// ::wxLogMessage(wxT("Assuming data sync"));
						m_radioSyncCounter = 0U;
						m_radioSyncsLost++;
						if (m_radioSyncsLost == MAX_LOST_SYNCS) {
							// ::wxLogMessage(wxT("Lost sync"));
							setRadioState(DSRXS_LISTENING);
						} else {
							processSlowData(true);
						}
					} else {
						m_radioSyncCounter++;
						CHeaderData* header = processSlowData(false);

						if (header != NULL) {
							bool res = processRadioHeader(header);
							if (res) {
								// A valid header and is a DV packet, go to normal data relaying
								setRadioState(DSRXS_PROCESS_DATA);
							} else {
								// This is a DD packet or some other problem
								// ::wxLogMessage(wxT("Invalid header"));
							}
						}
					}

					m_bitBuffer.clear();
				}
			}
			break;
	}
}

void CDStarRepeaterThread::repeaterStateMachine()
{
	switch (m_rptState) {
		case DSRS_VALID:
			if (m_timeoutTimer.isRunning() && m_timeoutTimer.hasExpired()) {
				::wxLogMessage(wxT("User has timed out"));
				setRepeaterState(DSRS_TIMEOUT);
			}
			break;

		case DSRS_VALID_WAIT:
			if (m_ackTimer.hasExpired()) {
				if (m_mode != MODE_GATEWAY)
					transmitStatus();
				setRepeaterState(DSRS_LISTENING);
				m_activeHangTimer.start();
			}
			break;

		case DSRS_INVALID_WAIT:
			if (m_ackTimer.hasExpired()) {
				if (m_mode != MODE_GATEWAY)
					transmitErrorStatus();
				setRepeaterState(DSRS_LISTENING);
				m_activeHangTimer.start();
			}
			break;

		case DSRS_TIMEOUT_WAIT:
			if (m_ackTimer.hasExpired()) {
				if (m_mode != MODE_GATEWAY)
					transmitStatus();
				setRepeaterState(DSRS_LISTENING);
				m_activeHangTimer.start();
			}
			break;

		case DSRS_NETWORK:
			if (m_watchdogTimer.hasExpired()) {
				::wxLogMessage(wxT("Network watchdog has expired"));
				// Send end of transmission data to the radio
				m_dataBuffer.addData(END_PATTERN_BITS, END_PATTERN_LENGTH_BITS);
				setRepeaterState(DSRS_LISTENING);
				m_activeHangTimer.start();
			}
			break;

		default:
			break;
	}
}

bool CDStarRepeaterThread::setRepeaterState(DSTAR_RPT_STATE state)
{
	// Can't change state when shutdown
	if (m_disable || m_shutdown)
		return false;

	// The "to" state
	switch (state) {
		case DSRS_SHUTDOWN:
			m_timeoutTimer.stop();
			m_watchdogTimer.stop();
			m_activeHangTimer.stop();
			m_ackTimer.stop();
			m_dataBuffer.clear();
			m_controller->setActive(false);
			m_controller->setRadioTransmit(false);
			m_rptState = DSRS_SHUTDOWN;
			break;

		case DSRS_LISTENING:
			m_timeoutTimer.stop();
			m_watchdogTimer.stop();
			m_ackTimer.stop();
			m_rptState = DSRS_LISTENING;
			if (m_protocolHandler != NULL)	// Tell the protocol handler
				m_protocolHandler->reset();
			break;

		case DSRS_VALID:
			if (m_rptState != DSRS_LISTENING)
				return false;

			m_activeHangTimer.stop();
			m_timeoutTimer.start();
			m_rptState = DSRS_VALID;
			break;

		case DSRS_VALID_WAIT:
			m_ackTimer.start();
			m_timeoutTimer.stop();
			m_rptState = DSRS_VALID_WAIT;
			break;

		case DSRS_TIMEOUT:
			// Send end of transmission data to the radio and the network
			m_dataBuffer.addData(END_PATTERN_BITS, END_PATTERN_LENGTH_BITS);
			if (m_protocolHandler != NULL) {
				unsigned char bytes[DV_FRAME_LENGTH_BYTES];
				::memcpy(bytes, NULL_AMBE_DATA_BYTES, VOICE_FRAME_LENGTH_BYTES);
				::memcpy(bytes + VOICE_FRAME_LENGTH_BYTES, NULL_SLOW_DATA_BYTES, DATA_FRAME_LENGTH_BYTES);
				m_protocolHandler->writeData(bytes, DV_FRAME_LENGTH_BYTES, true);
			}
			m_timeoutTimer.stop();
			m_rptState = DSRS_TIMEOUT;
			break;

		case DSRS_TIMEOUT_WAIT:
			m_ackTimer.start();
			m_timeoutTimer.stop();
			m_rptState = DSRS_TIMEOUT_WAIT;
			break;

		case DSRS_INVALID:
			if (m_rptState != DSRS_LISTENING)
				return false;

			m_activeHangTimer.stop();
			m_timeoutTimer.stop();
			m_rptState = DSRS_INVALID;
			break;

		case DSRS_INVALID_WAIT:
			m_ackTimer.start();
			m_timeoutTimer.stop();
			m_rptState = DSRS_INVALID_WAIT;
			break;

		case DSRS_NETWORK:
			if (m_rptState != DSRS_LISTENING)
				return false;

			m_rptState = DSRS_NETWORK;
			m_networkSyncCounter = 20U;
			m_activeHangTimer.stop();
			m_timeoutTimer.stop();
			m_watchdogTimer.start();
			m_ackTimer.stop();

			// Supress any received ack data
			delete m_statusHeader;
			m_statusHeader = NULL;
			break;
	}

	return true;
}

CHeaderData* CDStarRepeaterThread::processFECHeader()
{
	bool buffer[FEC_SECTION_LENGTH_BITS];
	m_bitBuffer.getData(buffer, FEC_SECTION_LENGTH_BITS);

	CRadioHeaderDecoder decoder(buffer, FEC_SECTION_LENGTH_BITS);

	CHeaderData* header = decoder.getHeaderData();
	if (header == NULL)
		return NULL;

	unsigned int ber = decoder.getBER();

	::wxLogMessage(wxT("Radio header decoded - My: %s/%s  Your: %s  Rpt1: %s  Rpt2: %s  Flags: %02X %02X %02X  BER: %u%%"), header->getMyCall1().c_str(), header->getMyCall2().c_str(), header->getYourCall().c_str(), header->getRptCall1().c_str(), header->getRptCall2().c_str(), header->getFlag1(), header->getFlag2(), header->getFlag3(), ber);

	return header;
}

bool CDStarRepeaterThread::processRadioHeader(CHeaderData* header)
{
	wxASSERT(header != NULL);

	// Check control messages
	bool res = checkControl(*header);
	if (res) {
		delete header;
		return true;
	}

	// If shutdown we ignore incoming headers
	if (m_rptState == DSRS_SHUTDOWN) {
		delete header;
		return true;
	}

	// Inform the gateway of all heard stations
	if (m_protocolHandler != NULL)
		m_protocolHandler->writeHeard(header->getMyCall1());

	// We don't handle DD data packets
	if (header->isDataPacket()) {
		::wxLogMessage(wxT("Received a DD packet, ignoring"));
		delete header;
		return false;
	}

	// The repeater bit must be set when not in gateway mode
	if (m_mode != MODE_GATEWAY) {
		if (!header->isRepeaterMode()) {
			::wxLogMessage(wxT("Received a non-repeater packet, ignoring"));

			bool res = setRepeaterState(DSRS_INVALID);
			if (res) {
				delete m_userHeader;
				m_userHeader = header;
			} else {
				delete header;
			}

			return true;
		}
	} else {
		// Quietly reject acks when in gateway mode
		if (header->isAck() || header->isNoResponse() || header->isRelayUnavailable()) {
			delete header;
			return true;
		}
	}

	// Make sure MyCall is not empty or the repeater or gateway callsigns
	if (header->getMyCall1().IsSameAs(m_rptCallsign) ||
		header->getMyCall1().IsSameAs(m_gwyCallsign) ||
		header->getMyCall1().IsSameAs(wxT("        "))) {
		::wxLogMessage(wxT("Invalid MYCALL value, ignoring"));
		delete header;
		return true;
	}

	// Is it for us?
	if (!header->getRptCall1().IsSameAs(m_rptCallsign)) {
		::wxLogMessage(wxT("Invalid RPT1 value, ignoring"));

		bool res = setRepeaterState(DSRS_INVALID);
		if (res) {
			delete m_userHeader;
			m_userHeader = header;
		} else {
			delete header;
		}

		return true;
	}

	// If RPT2 is the repeater callsign, remove it
	if (header->getRptCall2().IsSameAs(m_rptCallsign)) {
		::wxLogMessage(wxT("Invalid RPT2 value, setting to blanks"));
		header->setRptCall2(wxT("        "));
	}

	setRepeaterState(DSRS_VALID);

	// Send the valid header to the gateway if we are accepted
	if (m_rptState == DSRS_VALID) {
		delete m_userHeader;
		m_userHeader = header;

		// Only send on the network if we have one
		if (m_protocolHandler != NULL) {
			// Remove the repeater flag from the header and swap RPT1 and RPT2, otherwise send 'as is'
			CHeaderData networkHeader(*m_userHeader);
			networkHeader.setRptCall1(m_userHeader->getRptCall2());
			networkHeader.setRptCall2(m_userHeader->getRptCall1());
			networkHeader.setFlag1(m_userHeader->getFlag1() & ~REPEATER_MASK);
			m_protocolHandler->writeHeader(networkHeader);
		}

		// Create the new radio header and transmit it, but only in duplex mode
		if (m_mode == MODE_DUPLEX) {
			CHeaderData txHeader(*m_userHeader);
			txHeader.setRptCall1(m_rptCallsign);
			txHeader.setRptCall2(m_gwyCallsign);
			txHeader.setFlag1(m_userHeader->getFlag1() & ~REPEATER_MASK);
			transmitHeader(txHeader);
		}
	} else {
		delete header;
	}

	return true;
}

void CDStarRepeaterThread::processNetworkHeader(NETWORK_TYPE type, CHeaderData* header)
{
	wxASSERT(type == NETWORK_USER_HEADER || type == NETWORK_STATUS_HEADER);
	wxASSERT(header != NULL);

	// If shutdown we ignore incoming headers
	if (m_rptState == DSRS_SHUTDOWN) {
		delete header;
		return;
	}

	// We simply store a status header for later
	if (type == NETWORK_STATUS_HEADER) {
		delete m_statusHeader;
		m_statusHeader = header;
		::wxLogMessage(wxT("Network status header received - My: %s/%s  Your: %s  Rpt1: %s  Rpt2: %s  Flags: %02X %02X %02X"), m_statusHeader->getMyCall1().c_str(), m_statusHeader->getMyCall2().c_str(), m_statusHeader->getYourCall().c_str(), m_statusHeader->getRptCall1().c_str(), m_statusHeader->getRptCall2().c_str(), m_statusHeader->getFlag1(), m_statusHeader->getFlag2(), m_statusHeader->getFlag3());
		return;
	}

	::wxLogMessage(wxT("Network user header received - My: %s/%s  Your: %s  Rpt1: %s  Rpt2: %s  Flags: %02X %02X %02X"), header->getMyCall1().c_str(), header->getMyCall2().c_str(), header->getYourCall().c_str(), header->getRptCall1().c_str(), header->getRptCall2().c_str(), header->getFlag1(), header->getFlag2(), header->getFlag3());

	// We don't handle DD data packets
	if (header->isDataPacket()) {
		::wxLogMessage(wxT("Received a DD packet, ignoring"));
		delete header;
		return;
	}

	setRepeaterState(DSRS_NETWORK);

	// Transmit the header as received over the network, if accepted
	if (m_rptState == DSRS_NETWORK) {
		delete m_userHeader;
		m_userHeader = header;

		// If in gateway mode, set the repeater bit and change RPT1 just for transmission
		if (m_mode == MODE_GATEWAY) {
			CHeaderData txHeader(*m_userHeader);
			txHeader.setRepeaterMode(true);
			txHeader.setRptCall1(m_rptCallsign);
			transmitHeader(txHeader);
		} else {
			transmitHeader(*m_userHeader);
		}
	} else {
		delete header;
	}
}

void CDStarRepeaterThread::processRadioFrame(FRAME_TYPE type)
{
	bool bits[DV_FRAME_LENGTH_BITS];
	m_bitBuffer.getData(bits, DV_FRAME_LENGTH_BITS);

	// If a sync frame, regenerate the sync bytes (bits)
	if (type == FRAME_SYNC)
		::memcpy(bits + VOICE_FRAME_LENGTH_BITS, DATA_SYNC_BITS, DATA_FRAME_LENGTH_BITS * sizeof(bool));

	// Only regenerate the AMBE on received radio data
	if (type != FRAME_END && (m_rptState == DSRS_VALID || m_rptState == DSRS_INVALID || m_rptState == DSRS_TIMEOUT)) {
		m_ambeBits   += 48U;		// Only count the bits with FEC added
		m_ambeErrors += m_ambe.regenerate(bits);
	}

	// Don't pass through the frame of an invalid transmission
	if (m_rptState != DSRS_VALID)
		return;

	if (type == FRAME_END) {
		// Transmit the end sync on the radio, but only in duplex mode
		if (m_mode == MODE_DUPLEX) {
			m_dataBuffer.addData(END_PATTERN_BITS, END_PATTERN_LENGTH_BITS);
			m_dataBuffer.addData(END_PATTERN_BITS, END_PATTERN_LENGTH_BITS);
			m_dataBuffer.addData(END_PATTERN_BITS, END_PATTERN_LENGTH_BITS);
		}

		// Send null data and the end marker over the network
		if (m_protocolHandler != NULL) {
			unsigned char bytes[DV_FRAME_LENGTH_BYTES];
			::memcpy(bytes, NULL_AMBE_DATA_BYTES, VOICE_FRAME_LENGTH_BYTES);
			::memcpy(bytes + VOICE_FRAME_LENGTH_BYTES, NULL_SLOW_DATA_BYTES, DATA_FRAME_LENGTH_BYTES);
			m_protocolHandler->writeData(bytes, DV_FRAME_LENGTH_BYTES, true);
		}
	} else {
		// Convert the bits to bytes for the network
		unsigned char bytes[DV_FRAME_LENGTH_BYTES];
		unsigned int n = 0U;
		for (unsigned int i = 0U; i < DV_FRAME_LENGTH_BYTES; i++, n += 8U)
			bytes[i] = CUtils::bitsToByteRev(bits + n);

		// Send the data for transmission, but only in duplex mode
		if (m_mode == MODE_DUPLEX)
			m_dataBuffer.addData(bits, DV_FRAME_LENGTH_BITS);

		// Send the data to the network
		if (m_protocolHandler != NULL)
			m_protocolHandler->writeData(bytes, DV_FRAME_LENGTH_BYTES, false);
	}
}

void CDStarRepeaterThread::processNetworkFrame(NETWORK_TYPE type, unsigned char* data, unsigned int length, bool end)
{
	wxASSERT(type == NETWORK_USER_DATA || type == NETWORK_STATUS_DATA);
	wxASSERT(data != NULL);
	wxASSERT(length > 0U);

	// We do nothing with the data payload of statuses
	if (type == NETWORK_STATUS_DATA)
		return;

	if (m_rptState != DSRS_NETWORK)
		return;

	// Handle the sync bytes from the network
	if (::memcmp(data + VOICE_FRAME_LENGTH_BYTES, DATA_SYNC_BYTES, DATA_FRAME_LENGTH_BYTES) == 0) {
		m_networkSyncCounter = 0U;
	} else if (m_networkSyncCounter == 20U && !end) {
		// Regenerate the sync bytes
		::memcpy(data + VOICE_FRAME_LENGTH_BYTES, DATA_SYNC_BYTES, DATA_FRAME_LENGTH_BYTES);
		m_networkSyncCounter = 0U;
	} else {
		m_networkSyncCounter++;
	}

	if (!end) {
		m_ambe.regenerate(data);

		// Convert the bytes to bits for transmission
		bool bits[DV_FRAME_LENGTH_BITS * 2U];
		unsigned int n = 0U;
		for (unsigned int i = 0U; i < length; i++, n += 8U)
			CUtils::byteToBitsRev(data[i], bits + n);

		m_dataBuffer.addData(bits, length * 8U);
	} else {
		m_dataBuffer.addData(END_PATTERN_BITS, END_PATTERN_LENGTH_BITS);
		m_dataBuffer.addData(END_PATTERN_BITS, END_PATTERN_LENGTH_BITS);
		m_dataBuffer.addData(END_PATTERN_BITS, END_PATTERN_LENGTH_BITS);

		m_activeHangTimer.start();
		setRepeaterState(DSRS_LISTENING);
	}
}

CHeaderData* CDStarRepeaterThread::processSlowData(bool sync)
{
	bool bits[DV_FRAME_LENGTH_BITS];
	m_bitBuffer.getData(bits, DV_FRAME_LENGTH_BITS);

	if (sync) {
		m_slowData.sync();
		return NULL;
	}

	unsigned char bytes[DATA_FRAME_LENGTH_BYTES];

	// Convert the bits to bytes
	unsigned int n = 0U;
	for (unsigned int i = 0U; i < DATA_FRAME_LENGTH_BYTES; i++, n += 8U)
		bytes[i] = CUtils::bitsToByteRev(bits + VOICE_FRAME_LENGTH_BITS + n);

	m_slowData.addData(bytes);

	CHeaderData* header = m_slowData.getHeaderData();
	if (header == NULL)
		return NULL;

	::wxLogMessage(wxT("Radio header from slow data - My: %s/%s  Your: %s  Rpt1: %s  Rpt2: %s  Flags: %02X %02X %02X  BER: 0%%"), header->getMyCall1().c_str(), header->getMyCall2().c_str(), header->getYourCall().c_str(), header->getRptCall1().c_str(), header->getRptCall2().c_str(), header->getFlag1(), header->getFlag2(), header->getFlag3());

	return header;
}

void CDStarRepeaterThread::endOfRadioData()
{
	switch (m_rptState) {
		case DSRS_VALID:
			::wxLogMessage(wxT("AMBE for %s  BER: %u%%"), m_userHeader->getMyCall1().c_str(), (m_ambeErrors * 100U) / m_ambeBits);
			if (m_ack || m_mode == MODE_GATEWAY) {
				setRepeaterState(DSRS_VALID_WAIT);
			} else {
				setRepeaterState(DSRS_LISTENING);
				m_activeHangTimer.start();
			}
			break;

		case DSRS_INVALID:
			::wxLogMessage(wxT("AMBE for %s  BER: %u%%"), m_userHeader->getMyCall1().c_str(), (m_ambeErrors * 100U) / m_ambeBits);
			if (m_ack || m_mode == MODE_GATEWAY) {
				setRepeaterState(DSRS_INVALID_WAIT);
			} else {
				setRepeaterState(DSRS_LISTENING);
				m_activeHangTimer.start();
			}
			break;

		case DSRS_TIMEOUT:
			::wxLogMessage(wxT("AMBE for %s  BER: %u%%"), m_userHeader->getMyCall1().c_str(), (m_ambeErrors * 100U) / m_ambeBits);
			if (m_ack || m_mode == MODE_GATEWAY) {
				setRepeaterState(DSRS_TIMEOUT_WAIT);
			} else {
				setRepeaterState(DSRS_LISTENING);
				m_activeHangTimer.start();
			}
			break;

		default:
			break;
	}
}

void CDStarRepeaterThread::callback(const wxFloat32* input, wxFloat32* output, unsigned int nSamples, int id)
{
	::memset(output, 0x00, nSamples * sizeof(wxFloat32));

	if (!m_stopped) {
		readRadioData(input, nSamples);

		m_outBuffer.getData(output, nSamples);

		if (m_audioDelay != NULL)
			m_audioDelay->delay(output, nSamples);

		// Scale the output audio level
		for (unsigned int i = 0U; i < nSamples; i++)
			output[i] *= m_level;
	}
}

void CDStarRepeaterThread::setRadioState(DSTAR_RX_STATE state)
{
	// This is the too state
	switch (state) {
		case DSRXS_LISTENING:
			m_frameMatcher.reset();
			m_dataMatcher.reset();
			m_rxState = DSRXS_LISTENING;
			break;

		case DSRXS_PROCESS_HEADER:
			m_bitBuffer.setMaxLength(FEC_SECTION_LENGTH_BITS);
			m_rxState = DSRXS_PROCESS_HEADER;
			break;

		case DSRXS_PROCESS_DATA:
			m_bitBuffer.setMaxLength(DV_FRAME_LENGTH_BITS);
			m_endMatcher.reset();
			m_dataMatcher.reset();
			m_ambeBits   = 0U;
			m_ambeErrors = 0U;
			m_rxState = DSRXS_PROCESS_DATA;
			break;

		case DSRXS_PROCESS_SLOW_DATA:
			m_bitBuffer.setMaxLength(DV_FRAME_LENGTH_BITS);
			m_radioSyncCounter = 0U;
			m_radioSyncsLost   = 0U;
			m_slowData.reset();
			m_endMatcher.reset();
			m_dataMatcher.reset();
			m_rxState = DSRXS_PROCESS_SLOW_DATA;
			break;
	}
}

void CDStarRepeaterThread::readRadioData(const wxFloat32* input, unsigned int nSamples)
{
	wxASSERT(input != NULL);
	wxASSERT(nSamples > 0U);

	if (m_reader != NULL) {		// Are we reading from a WAV file?
		wxFloat32 buffer[DSTAR_RADIO_BLOCK_SIZE * 2U];
		unsigned int n = m_reader->read(buffer, nSamples);
		if (n == 0U) {				// EOF on the input WAV file?
			CWAVFileReader* reader = m_reader;
			m_reader = NULL;

			reader->close();
			delete reader;

			::wxLogInfo(wxT("End of the input file"));

			// Send data from the sound card as the file has ended
			m_inBuffer.addData(input, nSamples);

			return;
		}

		// Send data from the sound file
		m_inBuffer.addData(buffer, n);
	} else {
		// Send data from the sound card
		m_inBuffer.addData(input, nSamples);
	}
}

CDStarRepeaterStatusData* CDStarRepeaterThread::getStatus()
{
	if (m_rptState == DSRS_SHUTDOWN || m_rptState == DSRS_LISTENING)
		return new CDStarRepeaterStatusData(wxEmptyString, wxEmptyString, wxEmptyString, wxEmptyString, wxEmptyString,
				0x00, 0x00, 0x00, m_tx, m_rxState, m_rptState, m_timeoutTimer.getTimer(),
				m_timeoutTimer.getTimeout(), m_activeHangTimer.getTimer(), m_activeHangTimer.getTimeout());
	else
		return new CDStarRepeaterStatusData(m_userHeader->getMyCall1(), m_userHeader->getMyCall2(),
				m_userHeader->getYourCall(), m_userHeader->getRptCall1(), m_userHeader->getRptCall2(), 
				m_userHeader->getFlag1(), m_userHeader->getFlag2(), m_userHeader->getFlag3(), m_tx,
				m_rxState, m_rptState, m_timeoutTimer.getTimer(), m_timeoutTimer.getTimeout(),
				m_activeHangTimer.getTimer(), m_activeHangTimer.getTimeout());
}

bool CDStarRepeaterThread::isActive() const
{
	return m_rptState != DSRS_LISTENING;
}

void CDStarRepeaterThread::reset()
{
	m_rxState  = DSRXS_LISTENING;
	m_rptState = DSRS_LISTENING;

	m_timeoutTimer.stop();
	m_watchdogTimer.stop();
	m_activeHangTimer.stop();
	m_ackTimer.stop();
}

void CDStarRepeaterThread::clock()
{
	m_pollTimer.clock();
	m_timeoutTimer.clock();
	m_watchdogTimer.clock();
	m_activeHangTimer.clock();
	m_ackTimer.clock();
}

void CDStarRepeaterThread::shutdown()
{
	m_shutdown = true;
}

void CDStarRepeaterThread::startup()
{
	m_shutdown = false;
}

void CDStarRepeaterThread::command1()
{
	if (!m_controlCommand1Line.IsEmpty())
		::wxShell(m_controlCommand1Line);
}

void CDStarRepeaterThread::command2()
{
	if (!m_controlCommand2Line.IsEmpty())
		::wxShell(m_controlCommand2Line);
}

bool CDStarRepeaterThread::checkControl(const CHeaderData& header)
{
	if (!m_controlEnabled)
		return false;

	if (!m_controlRPT1.IsSameAs(header.getRptCall1()) || !m_controlRPT2.IsSameAs(header.getRptCall2()))
		return false;

	if (m_controlShutdown.IsSameAs(header.getYourCall())) {
		::wxLogMessage(wxT("Shutdown requested by %s/%s"), header.getMyCall1().c_str(), header.getMyCall2().c_str());
		shutdown();
	} else if (m_controlStartup.IsSameAs(header.getYourCall())) {
		::wxLogMessage(wxT("Startup requested by %s/%s"), header.getMyCall1().c_str(), header.getMyCall2().c_str());
		startup();
	} else if (m_controlCommand1.IsSameAs(header.getYourCall())) {
		::wxLogMessage(wxT("Command 1 requested by %s/%s"), header.getMyCall1().c_str(), header.getMyCall2().c_str());
		command1();
	} else if (m_controlCommand2.IsSameAs(header.getYourCall())) {
		::wxLogMessage(wxT("Command 2 requested by %s/%s"), header.getMyCall1().c_str(), header.getMyCall2().c_str());
		command2();
	} else if (m_controlOutput1.IsSameAs(header.getYourCall())) {
		::wxLogMessage(wxT("Output 1 requested by %s/%s"), header.getMyCall1().c_str(), header.getMyCall2().c_str());
		m_output1 = !m_output1;
		m_controller->setOutput1(m_output1);
	} else if (m_controlOutput2.IsSameAs(header.getYourCall())) {
		::wxLogMessage(wxT("Output 2 requested by %s/%s"), header.getMyCall1().c_str(), header.getMyCall2().c_str());
		m_output2 = !m_output2;
		m_controller->setOutput2(m_output2);
	} else if (m_controlOutput3.IsSameAs(header.getYourCall())) {
		::wxLogMessage(wxT("Output 3 requested by %s/%s"), header.getMyCall1().c_str(), header.getMyCall2().c_str());
		m_output3 = !m_output3;
		m_controller->setOutput3(m_output3);
	} else if (m_controlOutput4.IsSameAs(header.getYourCall())) {
		::wxLogMessage(wxT("Output 4 requested by %s/%s"), header.getMyCall1().c_str(), header.getMyCall2().c_str());
		m_output4 = !m_output4;
		m_controller->setOutput4(m_output4);
	} else {
		::wxLogMessage(wxT("Invalid command of %s sent by %s/%s"), header.getYourCall().c_str(), header.getMyCall1().c_str(), header.getMyCall2().c_str());
	}

	return true;
}
