/*
 *   Copyright (C) 2009 by Jonathan Naylor G4KLX
 *
 *   This program is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; either version 2 of the License, or
 *   (at your option) any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program; if not, write to the Free Software
 *   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include "ParrotControllerProtocolHandler.h"

#include "CCITTChecksum.h"
#include "DStarDefines.h"
#include "Utils.h"

const unsigned int BUFFER_LENGTH = 255U;

CParrotControllerProtocolHandler::CParrotControllerProtocolHandler(const wxString& repeaterAddress, unsigned int repeaterPort, const wxString& localAddress, unsigned int localPort) :
m_socket(repeaterAddress, repeaterPort, localAddress, localPort),
m_outId(0U),
m_outSeq1(0U),
m_outSeq2(0U),
m_type(NETWORK_NONE),
m_inId(0U),
m_inSeq1(0U),
m_userSeq2(0U),
m_statusSeq2(0U),
m_buffer(NULL),
m_length(0U)
{
	m_buffer = new unsigned char[BUFFER_LENGTH];

	wxDateTime now = wxDateTime::Now();
	::srand(now.GetMillisecond());
}

CParrotControllerProtocolHandler::~CParrotControllerProtocolHandler()
{
	delete[] m_buffer;
}

bool CParrotControllerProtocolHandler::open()
{
	return m_socket.open();
}

bool CParrotControllerProtocolHandler::writeHeader(const CHeaderData& header)
{
	unsigned char buffer[58U];

	buffer[0] = 'D';
	buffer[1] = 'S';
	buffer[2] = 'T';
	buffer[3] = 'R';

	m_outSeq1++;
	buffer[4] = m_outSeq1 / 256U;	// Packet sequence number
	buffer[5] = m_outSeq1 % 256U;

	buffer[6] = 0x73;				// Not a response
	buffer[7] = 0x12;				// Data type

	buffer[8] = 0x00;				// Length of 48 bytes following
	buffer[9] = 0x30;

	buffer[10] = 0x20;				// AMBE plus Slow Data following

	buffer[11] = 0x00;				// Parrot Controller to Repeater
	buffer[12] = 0x01;
	buffer[13] = 0x02;

	// Create a random id for this transmission
	m_outId = (::rand() % 65535U) + 1U;

	buffer[14] = m_outId / 256U;	// Unique session id
	buffer[15] = m_outId % 256U;

	buffer[16] = 0x80;

	buffer[17] = header.getFlag1();				// Flags 1, 2 and 3
	buffer[18] = header.getFlag2();
	buffer[19] = header.getFlag3();

	for (unsigned int i = 0U; i < LONG_CALLSIGN_LENGTH; i++)
		buffer[20 + i] = header.getRptCall1().GetChar(i);

	for (unsigned int i = 0U; i < LONG_CALLSIGN_LENGTH; i++)
		buffer[28 + i] = header.getRptCall2().GetChar(i);

	for (unsigned int i = 0U; i < LONG_CALLSIGN_LENGTH; i++)
		buffer[36 + i] = header.getYourCall().GetChar(i);

	for (unsigned int i = 0U; i < LONG_CALLSIGN_LENGTH; i++)
		buffer[44 + i] = header.getMyCall1().GetChar(i);

	for (unsigned int i = 0U; i < SHORT_CALLSIGN_LENGTH; i++)
		buffer[52 + i] = header.getMyCall2().GetChar(i);

	// Get the checksum for the header
	CCCITTChecksum csum;
	csum.update(buffer + 17, 4U * LONG_CALLSIGN_LENGTH + SHORT_CALLSIGN_LENGTH + 3U);
	csum.result(buffer + 56);

	m_outSeq2 = 0U;

#if defined(DUMP_TX)
	CUtils::dump(wxT("Sending Header"), buffer, 58U);
	return true;
#else
	return m_socket.write(buffer, 58U);
#endif
}

bool CParrotControllerProtocolHandler::writeData(const unsigned char* data, unsigned int length, bool end)
{
	wxASSERT(data != NULL);
	wxASSERT(length == DV_FRAME_LENGTH_BYTES);

	unsigned char buffer[29U];

	buffer[0] = 'D';
	buffer[1] = 'S';
	buffer[2] = 'T';
	buffer[3] = 'R';

	m_outSeq1++;
	buffer[4] = m_outSeq1 / 256U;	// Packet sequence number
	buffer[5] = m_outSeq1 % 256U;

	buffer[6] = 0x73;				// Request
	buffer[7] = 0x12;				// Data

	buffer[8] = 0x00;				// Length of 19 bytes following
	buffer[9] = 0x13;

	buffer[10] = 0x20;				// AMBE plus Slow Data following

	buffer[11] = 0x00;				// Parrot Controller to Repeater
	buffer[12] = 0x01;
	buffer[13] = 0x02;

	buffer[14] = m_outId / 256U;	// Unique session id
	buffer[15] = m_outId % 256U;

	// If this is a data sync, reset the sequence to zero
	if (data[9] == 0x55 && data[10] == 0x2D && data[11] == 0x16)
		m_outSeq2 = 0U;

	buffer[16] = m_outSeq2;
	m_outSeq2++;

	// Copy the AMBE data
	buffer[17] = data[0];
	buffer[18] = data[1];
	buffer[19] = data[2];
	buffer[20] = data[3];
	buffer[21] = data[4];
	buffer[22] = data[5];
	buffer[23] = data[6];
	buffer[24] = data[7];
	buffer[25] = data[8];

	buffer[26] = data[9];
	buffer[27] = data[10];
	buffer[28] = data[11];

	if (m_outSeq2 > 0x14U)
		m_outSeq2 = 0U;

#if defined(DUMP_TX)
	CUtils::dump(wxT("Sending Data"), buffer, 29U);
#else
	bool res = m_socket.write(buffer, 29U);
	if (!res)
		return false;
#endif

	if (!end)
		return true;

	buffer[0] = 'D';
	buffer[1] = 'S';
	buffer[2] = 'T';
	buffer[3] = 'R';

	m_outSeq1++;
	buffer[4] = m_outSeq1 / 256U;	// Packet sequence number
	buffer[5] = m_outSeq1 % 256U;

	buffer[6] = 0x73;				// Request
	buffer[7] = 0x12;				// Data

	buffer[8] = 0x00;				// Length of 19 bytes following
	buffer[9] = 0x13;

	buffer[10] = 0x20;				// AMBE plus Slow Data following

	buffer[11] = 0x00;				// Parrot Controller to Repeater
	buffer[12] = 0x01;
	buffer[13] = 0x02;

	buffer[14] = m_outId / 256U;	// Unique session id
	buffer[15] = m_outId % 256U;

	buffer[16] = m_outSeq2 | 0x40;	// End of data marker
	m_outSeq2++;

	buffer[17] = 0x00;
	buffer[18] = 0x00;
	buffer[19] = 0x00;
	buffer[20] = 0x00;
	buffer[21] = 0x00;
	buffer[22] = 0x00;
	buffer[23] = 0x00;
	buffer[24] = 0x00;
	buffer[25] = 0x00;

	buffer[26] = 0x00;
	buffer[27] = 0x00;
	buffer[28] = 0x00;

#if defined(DUMP_TX)
	CUtils::dump(wxT("Sending Data"), buffer, 29U);
	return true;
#else
	return m_socket.write(buffer, 29U);
#endif
}

NETWORK_TYPE CParrotControllerProtocolHandler::read()
{
	bool res = true;

	// Loop until we have no more data from the socket or we have data for the higher layers
	while (res)
		res = readPackets();

	return m_type;
}

bool CParrotControllerProtocolHandler::readPackets()
{
	m_type = NETWORK_NONE;

	// No more data?
	m_length = m_socket.read(m_buffer, BUFFER_LENGTH);
	if (m_length == 0U)
		return false;

	// Invalid packet type?
	if (m_buffer[0] != 'D' || m_buffer[1] != 'S' || m_buffer[2] != 'T' || m_buffer[3] != 'R')
		return true;

	// Heard or Poll data
	else if (m_buffer[6] == 0x73 && m_buffer[7] == 0x21)
		return true;

	// User data
	else if (m_buffer[6] == 0x73 && m_buffer[7] == 0x12 && m_buffer[10] == 0x20) {
		m_inSeq1 = m_buffer[4] * 256U + m_buffer[5];
		wxUint16 id = m_buffer[14] * 256U + m_buffer[15];

		if (m_buffer[16] == 0x80 && m_inId == 0U) {
			// Only process user headers with the first flag equal to 0x00 or 0x40
			if (m_buffer[17] == 0x00 || m_buffer[17] == 0x40) {
				m_inId = id;					// Take the stream id
				m_type = NETWORK_USER_HEADER;
				return false;
			} else {
				return true;
			}
		}

		// We're not listening for headers, throw it away
		if (m_buffer[16] == 0x80 && m_inId != 0U)
			return true;

		// Check that the stream id matches the valid header, reject otherwise
		if (id == m_inId) {
			// Is this the last packet in the stream?
			if ((m_buffer[16] & 0x40) == 0x40)
				m_inId = 0U;

			m_userSeq2 = m_buffer[16] & 0x0F;
			m_type = NETWORK_USER_DATA;
			return false;
		} else {
			return true;
		}
	}

	// An unknown type
	CUtils::dump(wxT("Unknown packet type from Repeater"), m_buffer, m_length);

	return true;
}

CHeaderData* CParrotControllerProtocolHandler::readHeader()
{
	if (m_type != NETWORK_USER_HEADER)
		return NULL;

	unsigned int dataLen = m_buffer[8] * 256U + m_buffer[9];

	// Header is the wrong length!
	if (dataLen != 48U) {		// RADIO_HEADER_LENGTH_BYTES + 7U
		CUtils::dump(wxT("Invalid header length from the Repeater"), m_buffer, dataLen + 17U);
		return NULL;
	}

	// Header checksum testing is enabled
	CHeaderData* header = new CHeaderData(m_buffer + 17, RADIO_HEADER_LENGTH_BYTES, true);

	if (!header->isValid()) {
		CUtils::dump(wxT("Header checksum failure from the Repeater"), m_buffer + 17, RADIO_HEADER_LENGTH_BYTES);
		delete header;
		return NULL;
	}

	return header;
}

unsigned int CParrotControllerProtocolHandler::readData(unsigned char* buffer, unsigned int length, bool& end)
{
	if (m_type != NETWORK_USER_DATA)
		return 0U;

	// Get the data length
	unsigned int dataLen = m_buffer[8] * 256U + m_buffer[9] - 7U;

	// Is our buffer too small?
	if (dataLen > length)
		dataLen = length;

	// Is it the end of the data?
	end = (m_buffer[16] & 0x40) == 0x40;

	::memcpy(buffer, m_buffer + 17, dataLen);

	return dataLen;
}

void CParrotControllerProtocolHandler::reset()
{
	m_inId = 0U;
}

void CParrotControllerProtocolHandler::close()
{
	m_socket.close();
}
