/*
 *   Copyright (C) 2009,2010 by Jonathan Naylor G4KLX
 *
 *   This program is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; either version 2 of the License, or
 *   (at your option) any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program; if not, write to the Free Software
 *   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include "ParrotControllerThread.h"
#include "ParrotControllerApp.h"
#include "DStarDefines.h"
#include "HeaderData.h"
#include "Utils.h"


CParrotControllerThread::CParrotControllerThread() :
wxThread(wxTHREAD_JOINABLE),
m_mode(PM_PARROT),
m_state(PS_WAITING),
m_beaconTimer(DSTAR_TICKS_PER_SEC, 0U),
m_beaconFileName(),
m_keepFile(false),
m_killed(false),
m_stopped(true),
m_protocolHandler(NULL),
m_header(NULL),
m_dataTimer(0U),
m_dataTimeout(0U),
m_turnaroundTimer(DSTAR_TICKS_PER_SEC, 2U),
m_watchdogTimer(DSTAR_TICKS_PER_SEC, 1U),
m_fileName(),
m_reader(),
m_writer()
{
}

CParrotControllerThread::~CParrotControllerThread()
{
}

void* CParrotControllerThread::Entry()
{
	// Wait here until we have the essentials to run
	while (!m_killed && m_protocolHandler == NULL)
		Sleep(500UL);		// 1/2 sec

	if (m_killed)
		return NULL;

	m_writer.setDirectory(::wxGetHomeDir());

	m_stopped = false;

	m_beaconTimer.start();

	::wxLogMessage(wxT("Starting the Parrot Controller thread"));

	while (!m_killed) {
		NETWORK_TYPE type;
		bool res;

		wxDateTime begin = wxDateTime::UNow();

		switch (m_mode) {
			case PM_PARROT:
				switch (m_state) {
					case PS_WAITING:
						type = m_protocolHandler->read();
						switch (type) {
							case NETWORK_NONE:
								if (m_turnaroundTimer.hasExpired()) {
									m_turnaroundTimer.stop();
									transmit(m_fileName);
								}
								break;
							case NETWORK_USER_HEADER:
								receive(type);
								break;
							default:
								break;
						}
						break;
					case PS_RECEIVING:
						type = m_protocolHandler->read();
						receive(type);
						break;
					case PS_TRANSMITTING:
						res = transmit();
						if (!res) {
							if (!m_keepFile) {
								::wxLogMessage(wxT("Deleting file %s"), m_fileName.c_str());
								::wxRemoveFile(m_fileName);
								m_fileName.Clear();
							}
						}
						break;
				}
				m_turnaroundTimer.clock();
				m_watchdogTimer.clock();
				break;

			case PM_BEACON:
				switch (m_state) {
					case PS_WAITING:
						if (m_beaconTimer.hasExpired()) {
							res = transmit(m_beaconFileName);
							if (res)
								m_beaconTimer.stop();
							else
								m_beaconTimer.start();
						}
						break;
					case PS_RECEIVING:
						break;
					case PS_TRANSMITTING:
						res = transmit();
						if (!res)
							m_beaconTimer.start();
						break;
				}
				m_beaconTimer.clock();
				break;
		}

		// Pause for the rest of the 20ms
		wxTimeSpan span = wxDateTime::UNow().Subtract(begin);
		long ms = span.GetMilliseconds().ToLong();
		if (ms < DSTAR_FRAME_TIME_MS)
			Sleep(DSTAR_FRAME_TIME_MS - ms);
	}

	::wxLogMessage(wxT("Stopping the Parrot Controller thread"));

	m_protocolHandler->close();
	delete m_protocolHandler;

	return NULL;
}

void CParrotControllerThread::kill()
{
	m_killed = true;
}

void CParrotControllerThread::setMode(PARROT_MODE mode, unsigned int beaconTime, const wxString& beaconFileName, unsigned int turnaroundTime, bool keepFile)
{
	if (!m_stopped)
		return;

	m_mode = mode;

	switch (m_mode) {
		case PM_PARROT:
			m_beaconTimer.setTimeout(0U);
			m_turnaroundTimer.setTimeout(turnaroundTime);
			m_beaconFileName = wxEmptyString;
			m_keepFile = keepFile;
			break;
		case PM_BEACON:
			m_beaconTimer.setTimeout(beaconTime);
			m_turnaroundTimer.setTimeout(0U);
			m_beaconFileName = beaconFileName;
			m_keepFile = true;
			break;
	}
}

void CParrotControllerThread::setNetwork(const wxString& repeaterAddress, unsigned int repeaterPort, const wxString& localAddress, unsigned int localPort)
{
	if (!m_stopped)
		return;

	m_protocolHandler = new CParrotControllerProtocolHandler(repeaterAddress, repeaterPort, localAddress, localPort);

	bool res = m_protocolHandler->open();
	if (!res) {
		delete m_protocolHandler;
		m_protocolHandler = NULL;
	}
}

CParrotControllerStatusData* CParrotControllerThread::getStatus()
{
	if (m_header == NULL)
		return new CParrotControllerStatusData(m_state, wxEmptyString, wxEmptyString, wxEmptyString, wxEmptyString,
				wxEmptyString, 0x00, 0x00, 0x00, m_dataTimer, m_dataTimeout, m_turnaroundTimer.getTimer(),
				m_turnaroundTimer.getTimeout(), m_beaconTimer.getTimer(), m_beaconTimer.getTimeout());
	else
		return new CParrotControllerStatusData(m_state, m_header->getMyCall1(), m_header->getMyCall2(),
				m_header->getYourCall(), m_header->getRptCall1(), m_header->getRptCall2(), m_header->getFlag1(),
				m_header->getFlag2(), m_header->getFlag3(), m_dataTimer, m_dataTimeout, m_turnaroundTimer.getTimer(),
				m_turnaroundTimer.getTimeout(), m_beaconTimer.getTimer(), m_beaconTimer.getTimeout());
}

bool CParrotControllerThread::receive(NETWORK_TYPE type)
{
	unsigned char buffer[255U];
	unsigned int length;
	bool res;
	bool end;

	switch (type) {
		case NETWORK_USER_HEADER:
			m_header = m_protocolHandler->readHeader();
			if (m_header == NULL)
				return false;
			::wxLogMessage(wxT("Radio header received - My: %s/%s  Your: %s  Rpt1: %s  Rpt2: %s  Flags: %02X %02X %02X"), m_header->getMyCall1().c_str(), m_header->getMyCall2().c_str(), m_header->getYourCall().c_str(), m_header->getRptCall1().c_str(), m_header->getRptCall2().c_str(), m_header->getFlag1(), m_header->getFlag2(), m_header->getFlag3());
			res = m_writer.open(*m_header);
			m_fileName = m_writer.getFileName();
			::wxLogMessage(wxT("Writing to %s"), m_fileName.c_str());
			m_watchdogTimer.start();
			m_dataTimeout = 0U;
			m_dataTimer   = 0U;
			m_state = PS_RECEIVING;
			break;

		case NETWORK_USER_DATA:
			length = m_protocolHandler->readData(buffer, 255U, end);
			if (length == 0U)
				return false;

			m_writer.write(buffer, length);
			m_dataTimeout++;
			m_watchdogTimer.reset();

			if (end) {
				m_writer.close();
				m_watchdogTimer.stop();
				m_turnaroundTimer.start();

				delete m_header;
				m_header = NULL;

				m_dataTimeout = 0U;
				m_state = PS_WAITING;

				return false;
			}
			break;

		default:
			if (m_watchdogTimer.hasExpired()) {
				::wxLogMessage(wxT("Network watchdog has expired"));

				m_writer.close();
				m_watchdogTimer.stop();
				m_turnaroundTimer.start();

				delete m_header;
				m_header = NULL;

				m_dataTimeout = 0U;
				m_state = PS_WAITING;

				return false;
			}
			break;
	}

	return true;
}

bool CParrotControllerThread::transmit(const wxString& fileName)
{
	::wxLogMessage(wxT("Transmitting file %s"), fileName.c_str());

	bool res = m_reader.open(fileName);
	if (!res) {
		::wxLogError(wxT("Couldn't open the file"));
		return false;
	}

	DVTFR_TYPE type = m_reader.read();
	if (type != DVTFR_HEADER) {
		::wxLogError(wxT("First record is not the header"));
		m_reader.close();
		return false;
	}

	m_header = m_reader.readHeader();
	if (m_header == NULL) {
		m_reader.close();
		return false;
	}

	::wxLogMessage(wxT("File header - My: %s/%s  Your: %s  Rpt1: %s  Rpt2: %s  Flags: %02X %02X %02X"), m_header->getMyCall1().c_str(), m_header->getMyCall2().c_str(), m_header->getYourCall().c_str(), m_header->getRptCall1().c_str(), m_header->getRptCall2().c_str(), m_header->getFlag1(), m_header->getFlag2(), m_header->getFlag3());

	res = m_protocolHandler->writeHeader(*m_header);
	if (!res) {
		m_reader.close();
		return false;
	}

	m_state = PS_TRANSMITTING;

	m_dataTimeout = m_reader.getRecords();
	m_dataTimer   = 0U;

	return true;
}

bool CParrotControllerThread::transmit()
{
	DVTFR_TYPE type = m_reader.read();
	if (type != DVTFR_DATA) {
		m_reader.close();
		return false;
	}

	unsigned char buffer[255U];
	bool end;
	unsigned int length = m_reader.readData(buffer, 255U, end);
	if (length == 0U) {
		m_reader.close();
		return false;
	}

	bool res = m_protocolHandler->writeData(buffer, length, end);
	if (!res) {
		m_reader.close();
		return false;
	}

	if (end) {
		m_dataTimer = 0U;
		m_dataTimeout = 0U;
		m_reader.close();

		m_state = PS_WAITING;
		delete m_header;
		m_header = NULL;

		return false;
	} else {
		m_dataTimer++;
		return true;
	}
}
